// Squidge's SMC Low-level access routines.  
// 
// Inspired and derived from routines provided by Samsung Electronics M/M R&D Center & FireFly. 
//
// 
#include "smc.h"

#include <string.h>
#include <stdlib.h>
#include "24x.h"


u32 smc_l2p[0x2000];
u32 smc_p2l[0x2000];
struct SMC_INFO smc_info;
extern unsigned long main_loop;

#define xsm_busy() while (!(rPDDAT & 0x200))
#define xsm_chip_en() rPDDAT &= (~0x80)
#define xsm_chip_dis() rPDDAT |= 0x80
#define xsm_cle_en() rPEDAT |= 0x20
#define xsm_cle_dis() rPEDAT &= (~0x20)
#define xsm_ale_en() rPEDAT |= 0x10
#define xsm_ale_dis() rPEDAT &= (~0x10)
#define xsm_wp_en() rPDDAT &= (~0x40)
#define xsm_wp_dis() rPDDAT |= 0x40
#define xsm_read_en() rPBCON &= 0xFFFF0000
#define xsm_read_dis() rPBCON = (rPBCON & 0xFFFF0000) | 0x5555
#define xsm_write_en() xsm_read_dis()
#define xsm_write_dis() xsm_read_en()

//------------------------------------------------------------------------------

static void xsm_write( u8 data)
{
  rPBDAT = (rPBDAT & 0xFF00) | data;
  rPEDAT &= (~0x08);
  rPEDAT |= 0x08;
}


//------------------------------------------------------------------------------

static u8 xsm_read()
{
  u8 data;
  
  rPDDAT &= (~0x100);
  data = rPBDAT & 0xFF;
  rPDDAT |= 0x100;
  return data;
}


//------------------------------------------------------------------------------

// assumes chip enabled
// bit 7: write protected = 0, write enabled = 1
// bit 6: busy = 0, ready = 1
// bit 0: success = 0, failed = 1
//
// returns 1 on success, 0 on fail
static u8 xsm_status()
{
  u32 i;
  u8 status;
  //
  xsm_cle_en();
  xsm_write_en();
  xsm_write(READ_STATUS_CMD);
  xsm_write_dis();
  xsm_cle_dis();
  //
  xsm_read_en();
  status = xsm_read();
  xsm_read_dis();
  //
  if (status == 0xC0)  // ready & write enabled
    return 1;
  else
    return 0;
}


//------------------------------------------------------------------------------
void xsmc_read_id( u8* buf, u32 length)
{
  
  u32 i;
  //
  xsm_chip_en();
  //
  xsm_cle_en();
  xsm_write_en();
  xsm_write(READ_ID_CMD);
  xsm_write_dis();
  xsm_cle_dis();
  //
  xsm_ale_en();
  xsm_write_en();
  xsm_write( 0);
  xsm_write_dis();
  xsm_ale_dis();
  //
  xsm_read_en();
  for (i=0;i<length;i++) *(buf+i) = xsm_read();
  xsm_read_dis();
  //
  xsm_chip_dis();
}

//------------------------------------------------------------------------------

// read an entire block.
u8 xsmc_read_block (u32 block, u8* buf, u32 mask)
{
	u32 page, i, lp, bit;
	
	// convert logical block to physical block, and then convert into page suitable for read1 command...
	
	if (smc_l2p[block] < LBA_UNUSED)
		page = smc_l2p[block] << 5;
	else
		return 0;
		
	// fill in buffer with 0x55 so we know if are missing sectors, etc
	memset (buf, 0x55, 32*512);
	
	xsm_chip_en();
	
	// setup command
	xsm_cle_en();
	xsm_write_en();
	xsm_write(READ1_CMD);
	xsm_write_dis();
	xsm_cle_dis();

	// We send the address and read each page individually, as it's faster than reading sequentially and then throwing away the 16
	// bytes between the pages (the spare area)
	bit = 1;
	for (lp = 0; lp < 32; lp++)
	{
		if (mask & bit) // sector needs to be read
		{
			xsm_chip_en();
	
			// address
			xsm_ale_en();
			xsm_write_en();
			xsm_write( 0x00);
			xsm_write( (u8)(page >> 0));
			xsm_write( (u8)(page >> 8));
			if (smc_info.mb >= 64) xsm_write( (u8)(page >> 16));
			xsm_write_dis();
			xsm_ale_dis();
	
			// wait for data to become available
			xsm_busy();
		
			// read data
			xsm_read_en();
			for (i = 0; i < 512; i++)
			{ 
				*buf = xsm_read();
				buf++;
			}
			xsm_read_dis();
			xsm_chip_dis(); // to abort the data transfer before the 16-byte spare area
		}
		else // waste of time reading the sector as its going to be written over shortly anyway
			buf += 512;
		bit <<= 1;
		page ++;
	}
	
	return 1;	
}

//------------------------------------------------------------------------------
void xsmc_read_spare( u32 page, u8* buf, u8 length)
{
  u32 i;

  //
  xsm_chip_en();
  //
  xsm_cle_en();
  xsm_read_dis();
  xsm_write(READ2_CMD);
  xsm_read_en();
  xsm_cle_dis();
  //
  xsm_ale_en();
  xsm_read_dis();
  xsm_write( 0x00);
  xsm_write( (u8)(page >> 0));
  xsm_write( (u8)(page >> 8));
  if (smc_info.mb >= 64) xsm_write( (u8)(page >> 16));
  xsm_read_en();
  xsm_ale_dis();
  //
  xsm_busy();
  //
  xsm_read_en();
  for (i=0;i<length;i++) *(buf+i) = xsm_read();
  xsm_read_dis();
  //
  xsm_chip_dis();
  //
}

// returns 1 on success, 0 on fail
static u8 xsmc_write_page_ext( u32 page, u8 *buf, u8 *spare )
{
  u32 i;
  u8 status;

  xsm_chip_en();
  //
  xsm_cle_en();
  xsm_write_en();
  xsm_write(RESET_PTR_CMD);
  xsm_write(SEQ_DATA_INPUT_CMD);
  xsm_write_dis();
  xsm_cle_dis();
  //
  xsm_ale_en();
  xsm_write_en();
  xsm_write( 0x00);
  xsm_write( (u8)(page >> 0));
  xsm_write( (u8)(page >> 8));
  if (smc_info.mb >= 64) xsm_write( (u8)(page >> 16));
  xsm_write_dis();
  xsm_ale_dis();
  //
  xsm_write_en();
  for (i = 0; i < 512; i++) xsm_write ( *(buf+i));
  for (i = 0; i < 16; i++) xsm_write ( *(spare+i)); 
  xsm_write_dis();
  //
  xsm_cle_en();
  xsm_write_en();
  xsm_write(PAGE_PROGRAM_CMD);
  xsm_write_dis();
  xsm_cle_dis();
  //
  xsm_busy();
  status = xsm_status();
  //
  xsm_chip_dis();
  return status;
}

//------------------------------------------------------------------------------
// returns 1 on success, 0 on fail
u8 xsmc_erase_block( u32 block)
{
  u8 status;
  u32 page;
  
  xsm_chip_en();
  //
  page = block * smc_info.pages_per_block;
  //
  xsm_cle_en();
  xsm_write_en();
  xsm_write(BLOCK_ERASE_CMD);
  xsm_write_dis();
  xsm_cle_dis();
  //
  xsm_ale_en();
  xsm_write_en();
  xsm_write( (u8)(page >> 0));
  xsm_write( (u8)(page >> 8));
  if (smc_info.mb >= 64) xsm_write( (u8)(page >> 16));
  xsm_write_dis();
  xsm_ale_dis();
  //
  xsm_cle_en();
  xsm_write_en();
  xsm_write(BLOCK_ERASE_CFM_CMD);
  xsm_write_dis();
  xsm_cle_dis();

  xsm_busy();
  status = xsm_status();
  
  xsm_chip_en();

  //
  return status;
}

// returns 1 on success, 0 on erase failure, 0xFF on write failure
static u8 xsmc_create_new_block (u32 new_pblock, u32 lblock, u8 *data)
{
	char str[256];
	u8 spare[16];
	u32 i;
	u32 old_pblock = smc_l2p[lblock];
	u32 zoned_lblock;
	u32 parity;

	if (old_pblock < BLOCK_UNUSED)
	{
		// step 3: we have found a block, so erase the old one (get physical of 'block') and mark as unused
		if (!xsmc_erase_block (old_pblock)) return 0;
		smc_p2l[old_pblock] = BLOCK_UNUSED;
	}

	// step 4: create zoned lblock and create new block
	zoned_lblock  = lblock % 1000;
	parity = 1;
	for (i = 0; i < 10; ++i) // zone converted lblocks are 10 bits
		parity ^= (zoned_lblock >> i) & 0x01;

	smc_l2p[lblock] = new_pblock;
	smc_p2l[new_pblock] = lblock;
	memset (spare, 0xFF, 16);
	spare[6] = spare[11] = (0x10 | ((zoned_lblock >> 7) & 0x07));
	spare[7] = spare[12] = (((zoned_lblock << 1) & 0xFE) | parity);
	
	// step 5: write our new block page by page, inserting our spare data at the end of each page
	for (i = 0; i < 32; i ++)
	{
		u8 *write_buffer;
		
		write_buffer = &data[i << 9];
		sm_ECCEncode((const u8 *)write_buffer, &spare[13]);
		sm_ECCEncode((const u8 *)write_buffer + 256, &spare[8]);
		if (!xsmc_write_page_ext ((new_pblock << 5) + i, write_buffer, spare)) return 0xFF; 						
	}
	
	return 1; 			
} 

// returns 1 on success, 0 on fail
u8 xsmc_write_block (u32 block, u8 *data)
{
	u32 pblock;
	u8 spare[16];
	u32 zone_start, zone_end, pblock_start;
	u8 zone;
	u8 status;
	u32 i;
	
	// split the smc up into 16mb chunks as 32mb and above was only thought about later on...
	zone = block / 1000;
	zone_start = zone * 1024;
	zone_end = zone_start + 1023;
	
	pblock_start = zone_start; // + (main_loop % 1024);  // find randomish block to start searching from to improve wear levelling
	if ((pblock_start == zone_start) && (zone == 0)) // make sure we don't even attempt to overwrite the card information area
		pblock_start ++;
		
	// step 1: find unused physical block in the same zone as the logical block, complain if none found
	for (pblock = pblock_start; pblock < zone_end; pblock ++)
	{
		if (smc_p2l[pblock] == BLOCK_UNUSED)
		{
			status = xsmc_create_new_block (pblock, block, data);
			if (status == 1)
				return 1;
			else
			{
				// either the old block or the new block is bad. Keep searching for a good block
				if (status == 0)
				{
					// could not delete the old block - smc will be corrupted if we continue (if it isn't already)
					smc_p2l[smc_l2p[block]] = BLOCK_RESERVED;
					smc_l2p[block] = BLOCK_RESERVED;
					return 0;
				}
				if (status == 0xFF) 
				{
					char spare[16];
					// write block failure. Try and attempt to find another. 
					xsmc_erase_block (pblock);
					for (i = 0; i < 16; i ++)
						spare[i] = 0x00;
					spare[5] = 0xF0; // mark as invalid block
					xsmc_write_page_ext (pblock << 5, data, spare); // nothing we can do if it fails. 
					smc_p2l[pblock] = BLOCK_RESERVED;
				}
			}
		}
	}	
	// Eeek! Couldn't find a block to occupy in the current zone! 
	return 0;
}

//------------------------------------------------------------------------------
// Create the physical <-> logical translation tables
void xsmc_make_l2p()
{
  u32 pblock, i, j, lblock, zone, count, cnt1, cnt2, cnt3;
  u8 data[512];
  //
  cnt1 = 0;
  cnt2 = 0;
  cnt3 = 0;
  //
  for (i=0;i<0x2000;i++)
  {
    smc_l2p[i] = LBA_RESERVED;
    smc_p2l[i] = BLOCK_RESERVED;
  }
  for (pblock=0;pblock<smc_info.blocks;pblock++)
  {
    // read physical block - first page
    xsmc_read_spare( pblock*smc_info.pages_per_block, (u8*)&data, 16);
    //
    zone = pblock >> 10; // divide by 1024 to get zone
    if ((data[5] == 0xFF) && ((data[6]&0xF8) == 0x10))
    {
      lblock = ((((data[6]<<8)|(data[7]<<0)) >> 1) & 0x03FF) + (zone * 1000);
      smc_l2p[lblock] = pblock;
      smc_p2l[pblock] = lblock;
      cnt1++;
    }
    else
    {
      count = 0;
      for (j=0;j<16;j++)
      {
        if (data[j] == 0xFF) count++;
      }
      if (count == 16)
      {
      	smc_p2l[pblock] = BLOCK_UNUSED;
      	cnt2++;
      }
      else
      {
      	smc_p2l[pblock] = BLOCK_RESERVED;
      	cnt3++;
      }	
    }
  }
}

/******************************************************************************/

void xsmc_detect( u8 id1, u8 id2, u8 id3)
{
  // init
  smc_info.id[0] = id1;
  smc_info.id[1] = id2;
  smc_info.id[2] = id3;
  smc_info.mb    = 0;
  smc_info.bytes_per_page  = 0;
  smc_info.pages_per_block = 0;
  smc_info.blocks          = 0;
  //
  switch (id1)
  {
    case SMC_SAMSUNG_ID:
    case SMC_TOSHIBA_ID:
    {
      switch (id2)
      {
        case SMC_16MB  : smc_info.mb = 16; break;
        case SMC_32MB  : smc_info.mb = 32; break;
        case SMC_64MB  : smc_info.mb = 64; break;
        case SMC_128MB : smc_info.mb = 128; break;
      }
      break;
    }
  }
  //
  switch (smc_info.mb)
  {
    case 16  : smc_info.bytes_per_page = 512; smc_info.pages_per_block = 32; smc_info.blocks = 0x0400; break;
    case 32  : smc_info.bytes_per_page = 512; smc_info.pages_per_block = 32; smc_info.blocks = 0x0800; break;
    case 64  : smc_info.bytes_per_page = 512; smc_info.pages_per_block = 32; smc_info.blocks = 0x1000; break;
    case 128 : smc_info.bytes_per_page = 512; smc_info.pages_per_block = 32; smc_info.blocks = 0x2000; break;
  }
}

//------------------------------------------------------------------------------
void xsmc_init( void)
{

}	

/**********
 * Function: sm_ECCEncode (completely ripped, unaltered, from the samsung routines)
 * Remark:
 *	- adopted from "ECC Algorithm for SmartMedia V3.0"
 *		by Memory Product & Technology, Samsung Electronics Co. (ecc30.pdf)
 **********/
int sm_ECCEncode(const u8 * p_buf, u8 * p_ecc) 
{
	u32 i, j;
	u8 paritr[256], tmp = 0, tmp2 = 0;
	u8 data_table0[16] = { 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0 };
	u8 data_table1[16] = { 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1 };
	u8 sum = 0, paritc = 0;
	u8 parit0c = 0, parit1c = 0, parit2c = 0, parit3c = 0;
	u8 parit4c = 0, parit5c = 0, parit6c = 0, parit7c = 0;
	u8 parit1_1, parit1_2, parit2_1, parit2_2, parit4_1, parit4_2;
	u8 parit8_1 = 0, parit8_2 = 0, parit16_1 = 0, parit16_2 = 0, parit32_1 = 0, parit32_2 = 0;
	u8 parit64_1 = 0, parit64_2 = 0, parit128_1 = 0, parit128_2 = 0, parit256_1 = 0, parit256_2 = 0;
	u8 parit512_1 = 0, parit512_2 = 0, parit1024_1 = 0, parit1024_2 = 0;
	u8* paritr_ptr;

	paritr_ptr = paritr;
	for (i = 0; i < 256; ++i, ++paritr_ptr, ++p_buf) 
	{
		paritc ^= *p_buf;
		tmp = (*p_buf & 0xf0) >> 4;
		tmp2 = *p_buf & 0x0f;

		switch (tmp) 
		{
			case 0:
			case 3:
			case 5:
			case 6:
			case 9:
			case 10:
			case 12:
			case 15:
				*paritr_ptr = *(data_table0 + tmp2);
				break;

			case 1:
			case 2:
			case 4:
			case 7:
			case 8:
			case 11:
			case 13:
			case 14:
				*paritr_ptr = *(data_table1 + tmp2);
				break;
		}
	}

	parit0c = (paritc & 0x01) ? 1 : 0;
	parit1c = (paritc & 0x02) ? 1 : 0;
	parit2c = (paritc & 0x04) ? 1 : 0;
	parit3c = (paritc & 0x08) ? 1 : 0;
	parit4c = (paritc & 0x10) ? 1 : 0;
	parit5c = (paritc & 0x20) ? 1 : 0;
	parit6c = (paritc & 0x40) ? 1 : 0;
	parit7c = (paritc & 0x80) ? 1 : 0;
	parit1_2 = parit6c ^ parit4c ^ parit2c ^ parit0c;
	parit1_1 = parit7c ^ parit5c ^ parit3c ^ parit1c;
	parit2_2 = parit5c ^ parit4c ^ parit1c ^ parit0c;
	parit2_1 = parit7c ^ parit6c ^ parit3c ^ parit2c;
	parit4_2 = parit3c ^ parit2c ^ parit1c ^ parit0c;
	parit4_1 = parit7c ^ parit6c ^ parit5c ^ parit4c;

	paritr_ptr = paritr;
	for (i = 0; i < 256; ++i, ++paritr_ptr) 
	{
		sum ^= *paritr_ptr;
	}

	paritr_ptr = paritr;
	for (i = 0; i < 256; i += 2, paritr_ptr += 2) 
	{
		parit8_2 ^= *paritr_ptr;
	}

	paritr_ptr = paritr;
	for (i = 0; i < 256; i += 4, paritr_ptr += 4) 
	{
		parit16_2 ^= *paritr_ptr;
		parit16_2 ^= *(paritr_ptr + 1);
	}

	paritr_ptr = paritr;
	for (i = 0; i < 256; i += 8, paritr_ptr += 8) 
	{
		for (j = 0; j <= 3; ++j) 
		{
			parit32_2 ^= *(paritr_ptr + j);
		}
	}

	paritr_ptr = paritr;
	for (i = 0; i < 256; i += 16, paritr_ptr += 16) 
	{
		for (j = 0; j <= 7; ++j) 
		{
			parit64_2 ^= *(paritr_ptr + j);
		}
	}

	paritr_ptr = paritr;
	for (i = 0; i < 256; i += 32, paritr_ptr += 32) 
	{
		for (j = 0; j <= 15; ++j) 
		{
			parit128_2 ^= *(paritr_ptr + j);
		}
	}

	paritr_ptr = paritr;
	for (i = 0; i < 256; i += 64, paritr_ptr += 64) 
	{
		for (j = 0; j <= 31; ++j) 
		{
			parit256_2 ^= *(paritr_ptr + j);
		}
	}

	paritr_ptr = paritr;
	for (i = 0; i < 256; i += 128, paritr_ptr += 128) 
	{
		for (j = 0; j <= 63; ++j) 
		{
			parit512_2 ^= *(paritr_ptr + j);
		}
	}

	paritr_ptr = paritr;
	for (i = 0; i < 256; i += 256, paritr_ptr += 256) 
	{
		for (j = 0; j <= 127; ++j) 
		{
			parit1024_2 ^= *(paritr_ptr + j);
		}
	}

	if (sum==0) 
	{
		parit1024_1 = parit1024_2;
		parit512_1 = parit512_2;
		parit256_1 = parit256_2;
		parit128_1 = parit128_2;
		parit64_1 = parit64_2;
		parit32_1 = parit32_2;
		parit16_1 = parit16_2;
		parit8_1 = parit8_2;
	}
	else 
	{
		parit1024_1 = parit1024_2 ? 0 : 1;
		parit512_1 = parit512_2 ? 0 : 1;
		parit256_1 = parit256_2 ? 0 : 1;
		parit128_1 = parit128_2 ? 0 : 1;
		parit64_1 = parit64_2 ? 0 : 1;
		parit32_1 = parit32_2 ? 0 : 1;
		parit16_1 = parit16_2 ? 0 : 1;
		parit8_1 = parit8_2 ? 0 : 1;
	}

	parit1_2 <<= 2;
	parit1_1 <<= 3;
	parit2_2 <<= 4;
	parit2_1 <<= 5;
	parit4_2 <<= 6;
	parit4_1 <<= 7;
	parit128_1 <<= 1;
	parit256_2 <<= 2;
	parit256_1 <<= 3;
	parit512_2 <<= 4;
	parit512_1 <<= 5;
	parit1024_2 <<= 6;
	parit1024_1 <<= 7;
	parit8_1 <<= 1;
	parit16_2 <<= 2;
	parit16_1 <<= 3;
	parit32_2 <<= 4;
	parit32_1 <<= 5;
	parit64_2 <<= 6;
	parit64_1 <<= 7;

	p_ecc[0] = ~(parit64_1 | parit64_2 | parit32_1 | parit32_2 | parit16_1 | parit16_2 | parit8_1 | parit8_2);
	p_ecc[1] = ~(parit1024_1 |parit1024_2 | parit512_1 | parit512_2 | parit256_1 | parit256_2 | parit128_1 | parit128_2);
	p_ecc[2] = ~(parit4_1 | parit4_2 | parit2_1 | parit2_2 | parit1_1 | parit1_2);

	return 0;
}
